=== modified file 'ChangeLog'
--- ChangeLog	2013-02-02 16:09:52 +0000
+++ ChangeLog	2013-02-08 16:15:23 +0000
@@ -1,3 +1,32 @@
+2013-02-08  James Hunt  <james.hunt@ubuntu.com>
+
+	* dbus/com.ubuntu.Upstart.xml: Added 'EndSession' method.
+	* init/Makefile.am: Updated for quiesce.[ch].
+	* init/conf.c: conf_destroy(): Cleanup function.
+	* init/conf.h: Prototype.
+	* init/control.c: control_end_session(): 'EndSession' implemenation.
+	* init/control.h: Include.
+	* init/events.h: Added SESSION_END_EVENT.
+	* init/job_class.c: job_class_max_kill_timeout(): New function.
+	* init/job_class.h: Prototype.
+	* init/job_process.c:
+	  - Added disable_respawn to disallow respawns.
+	  - job_process_jobs_running(): New function.
+	  - job_process_stop_all(): New function.
+	  - job_process_terminated(): Honours disable_respawn.
+	* init/job_process.h: Prototypes.
+	* init/main.c:
+	  - Typos.
+	  - term_handler(): Quiesce rather than re-exec on receipt of SIGTERM
+	    when running as a Session Init.
+	  - main(): Make quiesce() handle cleanup.
+	* init/man/init.8: Update for Session Init SIGTERM handling.
+	* init/man/startup.7: Update for Session Init.
+	* init/session.c: session_destroy(): New function.
+	* init/session.h: Prototype.
+	* init/man/session-end.7: New man page.
+	* init/quiesce.[ch]: New files.
+
 2013-02-02  James Hunt  <james.hunt@ubuntu.com>
 
 	* util/initctl.c: Remove ability to specify explicitly job and/or job

=== modified file 'dbus/com.ubuntu.Upstart.xml'
--- dbus/com.ubuntu.Upstart.xml	2013-01-25 20:08:49 +0000
+++ dbus/com.ubuntu.Upstart.xml	2013-02-08 16:15:23 +0000
@@ -103,6 +103,8 @@
     <method name="NotifyDiskWriteable">
     </method>
 
+    <method name="EndSession"/>
+
     <!-- Basic information about Upstart -->
     <property name="version" type="s" access="read" />
     <property name="log_priority" type="s" access="readwrite" />

=== modified file 'init/Makefile.am'
--- init/Makefile.am	2013-01-21 16:39:28 +0000
+++ init/Makefile.am	2013-02-08 16:15:23 +0000
@@ -31,7 +31,8 @@
 	man/keyboard-request.7 \
 	man/power-status-changed.7 \
 	man/upstart.7 \
-	man/inittab.5
+	man/inittab.5 \
+	man/session-end.7
 
 
 sbin_PROGRAMS = init
@@ -57,6 +58,7 @@
 	conf.c conf.h \
 	control.c control.h \
 	xdg.c xdg.h \
+	quiesce.c quiesce.h \
 	errors.h
 nodist_init_SOURCES = \
 	$(com_ubuntu_Upstart_OUTPUTS) \

=== modified file 'init/conf.c'
--- init/conf.c	2013-01-08 16:15:46 +0000
+++ init/conf.c	2013-02-08 16:15:23 +0000
@@ -321,6 +321,17 @@
 		conf_sources = NIH_MUST (nih_list_new (NULL));
 }
 
+/**
+ * conf_destroy:
+ *
+ * Clear: the conf_sources list.
+ **/
+void
+conf_destroy (void)
+{
+	if (conf_sources)
+		nih_free (conf_sources);
+}
 
 /**
  * conf_source_new:

=== modified file 'init/conf.h'
--- init/conf.h	2012-12-19 12:46:46 +0000
+++ init/conf.h	2013-02-08 16:15:23 +0000
@@ -112,6 +112,7 @@
 
 
 void        conf_init          (void);
+void        conf_destroy       (void);
 
 ConfSource *conf_source_new    (const void *parent, const char *path,
 				ConfSourceType type)

=== modified file 'init/control.c'
--- init/control.c	2013-01-31 17:23:55 +0000
+++ init/control.c	2013-02-08 16:15:23 +0000
@@ -57,6 +57,7 @@
 #include "errors.h"
 #include "state.h"
 #include "event.h"
+#include "events.h"
 #include "paths.h"
 #include "xdg.h"
 
@@ -115,9 +116,9 @@
 NihList *control_conns = NULL;
 
 /* External definitions */
-extern int user_mode;
-
-extern char *session_file;
+extern int      user_mode;
+extern int      disable_respawn;
+extern char    *session_file;
 
 /**
  * control_init:
@@ -1744,3 +1745,50 @@
 	if (session_file)
 		(void)unlink (session_file);
 }
+
+/**
+ * control_session_end:
+ *
+ * @data: not used,
+ * @message: D-Bus connection and message received.
+ *
+ * Implements the EndSession method of the com.ubuntu.Upstart
+ * interface.
+ *
+ * Called to request that Upstart stop all jobs and exit. Only
+ * appropriate when running as a Session Init and user wishes to
+ * 'logout'.
+ *
+ * Returns: zero on success, negative value on raised error.
+ **/
+int
+control_end_session (void             *data,
+		     NihDBusMessage   *message)
+{
+	Session  *session;
+
+	nih_assert (message);
+
+	/* Not supported at the system level */
+	if (getpid () == 1)
+		return 0;
+
+	if (! control_check_permission (message)) {
+		nih_dbus_error_raise_printf (
+			DBUS_INTERFACE_UPSTART ".Error.PermissionDenied",
+			_("You do not have permission to end session"));
+		return -1;
+	}
+
+	/* Get the relevant session */
+	session = session_from_dbus (NULL, message);
+
+	if (session && session->chroot) {
+		nih_warn (_("Ignoring session end request from chroot session"));
+		return 0;
+	}
+
+	quiesce (QUIESCE_REQUESTER_SESSION);
+
+	return 0;
+}

=== modified file 'init/control.h'
--- init/control.h	2013-01-31 17:23:55 +0000
+++ init/control.h	2013-02-08 16:15:23 +0000
@@ -31,6 +31,7 @@
 #include <json.h>
 
 #include "event.h"
+#include "quiesce.h"
 
 /**
  * USE_SESSION_BUS_ENV:

=== modified file 'init/events.h'
--- init/events.h	2009-06-23 09:29:35 +0000
+++ init/events.h	2013-02-08 16:15:23 +0000
@@ -31,6 +31,13 @@
 #define STARTUP_EVENT "debug"
 #endif
 
+/**
+ * SESSION_END_EVENT:
+ *
+ * Name of the event Upstart emits to denote a Session Init
+ * is shutting down.
+ **/
+#define SESSION_END_EVENT "session-end"
 
 /**
  * CTRLALTDEL_EVENT:

=== modified file 'init/job_class.c'
--- init/job_class.c	2013-01-31 17:23:55 +0000
+++ init/job_class.c	2013-02-08 16:15:23 +0000
@@ -2344,3 +2344,33 @@
 
 	return class;
 }
+
+/**
+ * job_class_max_kill_timeout:
+ *
+ * Determine maximum kill timeout for all running jobs.
+ *
+ * Returns: Maximum kill timeout (seconds).
+ **/
+time_t
+job_class_max_kill_timeout (void)
+{
+	time_t kill_timeout = JOB_DEFAULT_KILL_TIMEOUT;
+
+	job_class_init ();
+
+	NIH_HASH_FOREACH (job_classes, iter) {
+		JobClass *class = (JobClass *)iter;
+
+		NIH_HASH_FOREACH (class->instances, job_iter) {
+			Job *job = (Job *)job_iter;
+
+			if (job->class->kill_timeout > kill_timeout) {
+				kill_timeout = job->class->kill_timeout;
+				break;
+			}
+		}
+	}
+
+	return kill_timeout;
+}

=== modified file 'init/job_class.h'
--- init/job_class.h	2013-01-25 20:08:49 +0000
+++ init/job_class.h	2013-02-08 16:15:23 +0000
@@ -359,6 +359,9 @@
 JobClass * job_class_find (const Session *session, const char *name)
 	__attribute__ ((warn_unused_result));
 
+time_t     job_class_max_kill_timeout (void)
+	__attribute__ ((warn_unused_result));
+
 NIH_END_EXTERN
 
 #endif /* INIT_JOB_CLASS_H */

=== modified file 'init/job_process.c'
--- init/job_process.c	2013-01-25 09:08:09 +0000
+++ init/job_process.c	2013-02-08 16:15:23 +0000
@@ -99,6 +99,13 @@
  **/
 char *log_dir = NULL;
 
+/**
+ * disable_respawn:
+ *
+ * If TRUE, disallow respawning.
+ **/
+int disable_respawn = FALSE;
+
 /* Prototypes for static functions */
 static void job_process_error_abort     (int fd, JobProcessErrorType type,
 					 int arg)
@@ -128,8 +135,10 @@
 static void job_process_trace_fork      (Job *job, ProcessType process);
 static void job_process_trace_exec      (Job *job, ProcessType process);
 
-extern int          user_mode;
 extern char         *control_server_address;
+extern int           user_mode;
+extern int           session_end;
+extern time_t        quiesce_phase_time;
 
 /**
  * job_process_run:
@@ -1219,6 +1228,52 @@
 }
 
 /**
+ * job_process_jobs_running:
+ *
+ * Determine if any jobs are running.
+ *
+ * Returns: TRUE if jobs are still running, else FALSE.
+ **/
+int
+job_process_jobs_running (void)
+{
+	job_class_init ();
+
+	NIH_HASH_FOREACH (job_classes, iter) {
+		JobClass *class = (JobClass *)iter;
+
+		NIH_HASH_FOREACH (class->instances, job_iter)
+			return TRUE;
+	}
+
+	return FALSE;
+}
+
+
+/**
+ * job_process_stop_all:
+ *
+ * Stop all running jobs.
+ **/
+void
+job_process_stop_all (void)
+{
+	job_class_init ();
+
+	NIH_HASH_FOREACH (job_classes, iter) {
+		JobClass *class = (JobClass *)iter;
+
+		/* Note that instances get killed in a random order */
+		NIH_HASH_FOREACH (class->instances, job_iter) {
+			Job *job = (Job *)job_iter;
+
+			/* Request job instance stops */
+			job_change_goal (job, JOB_STOP);
+		}
+	}
+}
+
+/**
  * job_process_set_kill_timer:
  * @job: job to set kill timer for,
  * @process: process to be killed,
@@ -1547,7 +1602,7 @@
 			 * that's a simple matter of doing nothing.  Check
 			 * the job isn't running away first though.
 			 */
-			if (failed && job->class->respawn) {
+			if (failed && job->class->respawn && ! disable_respawn) {
 				if (job_process_catch_runaway (job)) {
 					nih_warn (_("%s respawning too fast, stopped"),
 						  job_name (job));

=== modified file 'init/job_process.h'
--- init/job_process.h	2012-11-26 19:30:41 +0000
+++ init/job_process.h	2013-02-08 16:15:23 +0000
@@ -63,7 +63,6 @@
 #define JOB_PROCESS_LOG_FILE_EXT ".log"
 #endif
 
-
 /**
  * JobProcessErrorType:
  *
@@ -151,6 +150,10 @@
 
 void   job_process_adj_kill_timer  (Job *job, time_t due);
 
+int    job_process_jobs_running (void);
+
+void   job_process_stop_all (void);
+
 NIH_END_EXTERN
 
 #endif /* INIT_JOB_PROCESS_H */

=== modified file 'init/main.c'
--- init/main.c	2013-01-31 17:23:55 +0000
+++ init/main.c	2013-02-08 16:15:23 +0000
@@ -431,9 +431,10 @@
 
 	}
 
-	/* SIGTERM instructs us to re-exec ourselves; this should be the
-	 * last in the list to ensure that all other signals are handled
-	 * before a SIGTERM.
+	/* SIGTERM instructs us to re-exec ourselves when running as PID
+	 * 1, or to exit when running as a Session Init; this signal should
+	 * be the last in the list to ensure that all other signals are
+	 * handled before a SIGTERM.
 	 */
 	nih_signal_set_handler (SIGTERM, nih_signal_handler);
 	NIH_MUST (nih_signal_add_handler (NULL, SIGTERM, term_handler, NULL));
@@ -653,8 +654,6 @@
 	nih_main_loop_interrupt ();
 	ret = nih_main_loop ();
 
-	control_cleanup ();
-
 	return ret;
 }
 
@@ -802,7 +801,8 @@
  * @signal: signal caught.
  *
  * This is called when we receive the TERM signal, which instructs us
- * to reexec ourselves.
+ * to reexec ourselves when running as PID 1, or to perform a controlled
+ * exit when running as a Session Init.
  **/
 static void
 term_handler (void      *data,
@@ -811,8 +811,12 @@
 	nih_assert (args_copy[0] != NULL);
 	nih_assert (signal != NULL);
 
+	if (user_mode) {
+		quiesce (QUIESCE_REQUESTER_SYSTEM);
+		return;
+	}
+
 	nih_warn (_("Re-executing %s"), args_copy[0]);
-
 	stateful_reexec ();
 }
 
@@ -823,7 +827,7 @@
  * @data: unused,
  * @signal: signal that called this handler.
  *
- * Handle having recieved the SIGINT signal, sent to us when somebody
+ * Handle having received the SIGINT signal, sent to us when somebody
  * presses Ctrl-Alt-Delete on the console.  We just generate a
  * ctrlaltdel event.
  **/
@@ -839,7 +843,7 @@
  * @data: unused,
  * @signal: signal that called this handler.
  *
- * Handle having recieved the SIGWINCH signal, sent to us when somebody
+ * Handle having received the SIGWINCH signal, sent to us when somebody
  * presses Alt-UpArrow on the console.  We just generate a
  * kbdrequest event.
  **/
@@ -855,7 +859,7 @@
  * @data: unused,
  * @signal: signal that called this handler.
  *
- * Handle having recieved the SIGPWR signal, sent to us when powstatd
+ * Handle having received the SIGPWR signal, sent to us when powstatd
  * changes the /etc/powerstatus file.  We just generate a
  * power-status-changed event and jobs read the file.
  **/
@@ -871,7 +875,7 @@
  * @data: unused,
  * @signal: signal that called this handler.
  *
- * Handle having recieved the SIGHUP signal, which we use to instruct us to
+ * Handle having received the SIGHUP signal, which we use to instruct us to
  * reload our configuration.
  **/
 static void
@@ -887,7 +891,7 @@
  * @data: unused,
  * @signal: signal that called this handler.
  *
- * Handle having recieved the SIGUSR signal, which we use to instruct us to
+ * Handle having received the SIGUSR signal, which we use to instruct us to
  * reconnect to D-Bus.
  **/
 static void

=== modified file 'init/man/init.8'
--- init/man/init.8	2013-01-18 15:26:17 +0000
+++ init/man/init.8	2013-02-08 16:15:23 +0000
@@ -144,8 +144,25 @@
 and pass all arguments to that.  See that manual page for further
 details. However, if the
 .B \-\-user
-option is specified, it will read alternative configuration files and
-manage the individual user session in a similar fashion.
+option is specified, it will run as a
+.B Session Init
+and read alternative configuration files and manage the individual user
+session in a similar fashion.
+
+Sending a Session Init a
+.I SIGTERM
+signal is taken as a request to shutdown due to an impending system
+shutdown. In this scenario, the
+Session Init will emit the 
+.B session\-end
+event and  request all running jobs stop. It will attempt to honour jobs
+.B kill timeout
+values (see
+.BR init (5)
+for further details). Note however that system policy will prevail: if
+jobs request timeout values longer than the system policy allows for
+complete system shutdown, it will not be possible to honour them before
+the Session Init is killed by the system.
 .\"
 .SH ENVIRONMENT VARIABLES
 

=== added file 'init/man/session-end.7'
--- init/man/session-end.7	1970-01-01 00:00:00 +0000
+++ init/man/session-end.7	2013-02-08 16:15:23 +0000
@@ -0,0 +1,34 @@
+.TH session\-end 7 2013-02-08 "Upstart"
+.\"
+.SH NAME
+session\-end \- event signalling session shutdown
+.\"
+.SH SYNOPSIS
+.B session\-end
+.BI TYPE\fR= TYPE
+.\"
+.SH DESCRIPTION
+The
+.B session\-end
+event is generated by the Upstart
+.BR init (8)
+daemon when running as a Session Init (where its pid is not 1) when
+requested to stop.
+
+The
+.B TYPE
+environment variable will be set to either
+.I shutdown
+when the Session Init has received the
+.I SIGTERM
+signal, or
+.I logout
+when the Session Init has processed the
+.B com.ubuntu.Upstart0_6.EndSession()
+D\-Bus method call.
+
+.\"
+.SH SEE ALSO
+.BR startup (7)
+.BR init (5)
+.BR init (8)

=== modified file 'init/man/startup.7'
--- init/man/startup.7	2009-07-09 12:26:11 +0000
+++ init/man/startup.7	2013-02-08 16:15:23 +0000
@@ -27,12 +27,17 @@
 .BR runlevel (7)
 event.  See that page for a more detailed explanation of this process.
 
+This event is emitted when running both as pid 1 and as a Session Init.
+
 Paradoxically there is currently no corresponding Upstart-native event
 signifying that the system is to be shutdown, only the System V compatible
 .B runlevel 0
 and
 .B runlevel 6
-events provide this functionality.
+events provide this functionality. However, when running as a Session
+Init, the
+.BR session\-end
+event performs this function.
 .\"
 .SH EXAMPLE
 A service with no other dependencies run on startup might use:
@@ -45,4 +50,6 @@
 .\"
 .SH SEE ALSO
 .BR runlevel (7)
+.BR session\-end (7)
+.BR init (5)
 .BR init (8)

=== added file 'init/quiesce.c'
--- init/quiesce.c	1970-01-01 00:00:00 +0000
+++ init/quiesce.c	2013-02-08 16:15:23 +0000
@@ -0,0 +1,248 @@
+/* upstart
+ *
+ * quiesce.c - shutdown handling.
+ *
+ * Copyright © 2013 Canonical Ltd.
+ * Author: James Hunt <james.hunt@canonical.com>
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License version 2, as
+ * published by the Free Software Foundation.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License along
+ * with this program; if not, write to the Free Software Foundation, Inc.,
+ * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
+ */
+
+#ifdef HAVE_CONFIG_H
+# include <config.h>
+#endif /* HAVE_CONFIG_H */
+
+#include "quiesce.h"
+#include "events.h"
+#include "environ.h"
+#include "conf.h"
+#include "job_process.h"
+#include "control.h"
+
+#include <nih/main.h>
+
+/**
+ * quiesce_requester:
+ *
+ * Where the quiesce request originated. This determines
+ * the shutdown behaviour.
+ **/
+static QuiesceRequester quiesce_requester = QUIESCE_REQUESTER_INVALID;
+
+/**
+ * quiesce_phase:
+ *
+ * Current phase of shutdown.
+ **/
+static QuiescePhase quiesce_phase = QUIESCE_PHASE_NOT_QUIESCED;
+
+/**
+ * max_kill_timeout:
+ *
+ * Maxiumum kill_timout value calculated from all running jobs used to
+ * determine how long to wait before exiting.
+ **/
+static time_t max_kill_timeout = 0;
+
+/**
+ * quiesce_phase_time:
+ *
+ * Time that a particular phase started.
+ **/
+static time_t quiesce_phase_time = 0;
+
+/* External definitions */
+extern int disable_respawn;
+
+/**
+ * quiesce:
+ *
+ * @requester: where the quiesce request originated.
+ *
+ * Commence Session Init shutdown.
+ **/
+void
+quiesce (QuiesceRequester requester)
+{
+	nih_local char  **env = NULL;
+	const char       *reason;
+
+	job_class_init ();
+
+	/* Quiesce already in progress */
+	if (quiesce_phase != QUIESCE_PHASE_NOT_QUIESCED)
+		return;
+
+	quiesce_requester = requester;
+
+	/* System shutdown skips the wait phase to ensure all running
+	 * jobs get signalled.
+	 *
+	 * Note that jobs which choose to start on SESSION_END_EVENT may
+	 * not complete (or even start), but no guarantee is possible in
+	 * the system shutdown scenario since Session Inits must not
+	 * hold up the system.
+	 */
+	quiesce_phase = (requester == QUIESCE_REQUESTER_SYSTEM)
+		? QUIESCE_PHASE_KILL
+		: QUIESCE_PHASE_WAIT;
+
+	reason = (requester == QUIESCE_REQUESTER_SESSION)
+		? _("logout") : _("shutdown");
+
+	nih_info (_("Quiescing due to %s request"), reason);
+
+	quiesce_phase_time = time (NULL);
+
+	/* Stop existing jobs from respawning */
+	disable_respawn = TRUE;
+
+	/* Signal that the session is ending. This may start new jobs.
+	 *
+	 * Note that the event doesn't actually get emitted until the
+	 * next time the main loop gets a chance to run.
+	 */
+	env = NIH_MUST (nih_str_array_new (NULL));
+
+	NIH_MUST (environ_set (&env, NULL, NULL, TRUE,
+				"TYPE=%s", reason));
+
+	NIH_MUST (event_new (NULL, SESSION_END_EVENT, env));
+
+	if (requester == QUIESCE_REQUESTER_SYSTEM) {
+		/* We'll attempt to wait for this long, but system
+		 * policy may prevent it such that we just get killed
+		 * and job processes reparented to PID 1.
+		 */
+		max_kill_timeout = job_class_max_kill_timeout ();
+
+		job_process_stop_all ();
+	}
+
+	/* Check every second to see if all jobs have finished. If so,
+	 * we can exit early.
+	 */
+	NIH_MUST (nih_timer_add_periodic (NULL, 1,
+				(NihTimerCb)quiesce_wait_callback, NULL));
+}
+
+/**
+ * quiesce_wait_callback:
+ *
+ * @data: not used,
+ * @timer: timer that caused us to be called.
+ *
+ * Callback used to check if all jobs have finished and if so
+ * finalise Session Init shutdown.
+ **/
+void
+quiesce_wait_callback (void *data, NihTimer *timer)
+{
+	time_t now;
+
+	nih_assert (timer);
+	nih_assert (quiesce_phase_time);
+
+	now = time (NULL);
+
+	nih_assert (quiesce_requester != QUIESCE_REQUESTER_INVALID);
+
+	if (quiesce_requester == QUIESCE_REQUESTER_SYSTEM) {
+		nih_assert (quiesce_phase == QUIESCE_PHASE_KILL);
+
+		if ((now - quiesce_phase_time) > max_kill_timeout)
+			goto out;
+
+	} else if (quiesce_phase == QUIESCE_PHASE_WAIT) {
+
+		if ((now - quiesce_phase_time) > QUIESCE_DEFAULT_JOB_RUNTIME) {
+			quiesce_phase = QUIESCE_PHASE_KILL;
+
+			/* reset for new phase */
+			quiesce_phase_time = time (NULL);
+
+			max_kill_timeout = job_class_max_kill_timeout ();
+			job_process_stop_all ();
+		}
+	} else if (quiesce_phase == QUIESCE_PHASE_KILL) {
+
+		if ((now - quiesce_phase_time) > max_kill_timeout)
+			goto out;
+	} else {
+		nih_assert_not_reached ();
+	}
+
+	if (! job_process_jobs_running ())
+		goto out;
+
+	return;
+
+out:
+	quiesce_show_slow_jobs ();
+
+	/* Note that we might skip the kill phase for the session
+	 * requestor if no jobs are actually running at this point.
+	 */
+	quiesce_phase = QUIESCE_PHASE_CLEANUP;
+	quiesce_finalise ();
+
+	/* Deregister */
+	nih_free (timer);
+}
+
+/**
+ * quiesce_show_slow_jobs:
+ *
+ * List jobs that are still running after their expected end time.
+ **/
+void
+quiesce_show_slow_jobs (void)
+{
+	job_class_init ();
+
+	NIH_HASH_FOREACH (job_classes, iter) {
+		JobClass *class = (JobClass *)iter;
+
+		/* Note that instances get killed in a random order */
+		NIH_HASH_FOREACH (class->instances, job_iter) {
+			nih_local const char  *name = NULL;
+			Job                   *job;
+		       
+			job = (Job *)job_iter;
+
+			name = job_name (job);
+
+			nih_message ("job %s failed to stop", name);
+		}
+	}
+}
+
+
+/**
+ * quiesce_finalise:
+ *
+ * Perform final shutdown operations.
+ **/
+void
+quiesce_finalise (void)
+{
+	nih_assert (quiesce_phase == QUIESCE_PHASE_CLEANUP);
+
+	/* Cleanup */
+	conf_destroy ();
+	session_destroy ();
+	control_cleanup ();
+
+	nih_main_loop_exit (0);
+}

=== added file 'init/quiesce.h'
--- init/quiesce.h	1970-01-01 00:00:00 +0000
+++ init/quiesce.h	2013-02-08 16:15:23 +0000
@@ -0,0 +1,76 @@
+/* upstart
+ *
+ * Copyright © 2013 Canonical Ltd.
+ * Author: James Hunt <james.hunt@canonical.com>
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License version 2, as
+ * published by the Free Software Foundation.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License along
+ * with this program; if not, write to the Free Software Foundation, Inc.,
+ * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
+ */
+
+#ifndef INIT_QUIESCE_H
+#define INIT_QUIESCE_H
+
+#include <nih/timer.h>
+
+/**
+ * QUIESCE_DEFAULT_JOB_RUNTIME:
+ *
+ * The default maximum length of time to wait after emitting the
+ * SESSION_END_EVENT event before stopping all jobs.
+ **/
+#define QUIESCE_DEFAULT_JOB_RUNTIME 5
+
+/**
+ * QuiesceRequester:
+ *
+ * Reason for Session Init wishing to shutdown; either the Session Init
+ * has been notified the system is being shutdown, or the session has
+ * requested it be ended (for example due to a user logout request).
+ **/
+typedef enum quiesce_requester {
+	QUIESCE_REQUESTER_INVALID = -1,
+	QUIESCE_REQUESTER_SYSTEM = 0,
+	QUIESCE_REQUESTER_SESSION,
+} QuiesceRequester;
+
+/**
+ * QuiescePhase:
+ *
+ * Phase 0: No quiesce operation in progress.
+ *
+ * Phase 1: Wait: Period between SESSION_END_EVENT being emitted and
+ *          QUIESCE_DEFAULT_JOB_RUNTIME being reached.
+ *
+ * Phase 2: Kill: Period between QUIESCE_DEFAULT_JOB_RUNTIME being
+ *          reached and kill signal being sent to all jobs.
+ *
+ * Phase 3: Cleanup: Period between all jobs having ended
+ *          (either naturally or by induction) and final exit.
+ **/
+typedef enum quiesce_phase {
+	QUIESCE_PHASE_NOT_QUIESCED,
+	QUIESCE_PHASE_WAIT,
+	QUIESCE_PHASE_KILL,
+	QUIESCE_PHASE_CLEANUP,
+} QuiescePhase;
+
+NIH_BEGIN_EXTERN
+
+void    quiesce                (QuiesceRequester requester);
+void    quiesce_wait_callback  (void *data, NihTimer *timer);
+void    quiesce_show_slow_jobs (void);
+void    quiesce_finalise       (void);
+
+NIH_END_EXTERN
+
+#endif /* INIT_QUIESCE_H */

=== modified file 'init/session.c'
--- init/session.c	2013-01-25 09:01:00 +0000
+++ init/session.c	2013-02-08 16:15:23 +0000
@@ -89,6 +89,18 @@
 		sessions = NIH_MUST (nih_list_new (NULL));
 }
 
+/**
+ * session_destroy:
+ *
+ * Clean up the sessions list.
+ **/
+void
+session_destroy (void)
+{
+	if (sessions)
+		nih_free (sessions);
+}
+
 
 /**
  * session_new:

=== modified file 'init/session.h'
--- init/session.h	2013-01-25 09:01:00 +0000
+++ init/session.h	2013-02-08 16:15:23 +0000
@@ -78,6 +78,7 @@
 extern NihList *sessions;
 
 void           session_init        (void);
+void           session_destroy     (void);
 
 Session      * session_new         (const void *parent, const char *chroot)
 	__attribute__ ((malloc, warn_unused_result));

